/*
 * ras.l			2001/04/15
 *
 * ras - Mobile Robot Assembler (MC68HC11)
 * Scanner code
 *
 * Copyright (C) 2001 Javier de Lope Asian <jdlope@eui.upm.es>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

%x incl
%option noyywrap

%{
#include <string.h>	// strcpy
#include <ctype.h>	// toupper

#define MAX_INCLUDE_DEPTH 10

struct incl_info {
  YY_BUFFER_STATE state;
  int             line;
  char		  name[256];
};

struct incl_info incl_stack[MAX_INCLUDE_DEPTH];

extern char fname[256];
int incl_stack_ptr = 0;
%}

		int lineno = 1;

decimal		[0-9]+
hexadecimal	$[0-9A-Fa-f]+
octal		@[0-7]+
binario		%[01]+
ascii		'.+

hexerr		$[0-9A-Za-z]+
octerr		@[0-9]+
binerr		%[0-9]+

colon		\:
comma		,
inmediato	#
regx		,X
regy		,Y

simbolo		[A-Za-z][A-Za-z0-9_]+

%%

include		|
INCLUDE		BEGIN(incl);

<incl>[ \t]*	;

<incl>[^ \t\n]+ { char str[128];
		  if (incl_stack_ptr >= MAX_INCLUDE_DEPTH) {
		    show_err(lineno, "profundidad de include superada");
		    exit(1);
		  }
		  incl_stack[incl_stack_ptr].state = YY_CURRENT_BUFFER;
		  incl_stack[incl_stack_ptr].line  = lineno;
		  strcpy(incl_stack[incl_stack_ptr].name, fname);
		  incl_stack_ptr++;
		  if (! (yyin = fopen(yytext, "r")) ) {
		    sprintf(str, "no se puede abrir el fichero `%s'", yytext);
		    show_err(lineno, str);
		    exit(1);
		  }
		  lineno = 1;
		  strcpy(fname, yytext);
		  yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
		  BEGIN(INITIAL);
		}

<<EOF>>		{ if (--incl_stack_ptr < 0)
		    yyterminate();
		  else {
		    yy_delete_buffer(YY_CURRENT_BUFFER);
		    yy_switch_to_buffer(incl_stack[incl_stack_ptr].state);
		    lineno = incl_stack[incl_stack_ptr].line;
		    strcpy(fname, incl_stack[incl_stack_ptr].name);
		    return (FL);
		  }
		}

end		|
END		{ return (END); }

equ		|
EQU		{ return (EQU); }

org		|
ORG		{ return (ORG); }

fcb		|
FCB		{ return (FCB); }

fdb		|
FDB		{ return (FDB); }

fcc		|
FCC		{ return (FCC); }

rmb		|
RMB		{ return (RMB); }

zmb		|
ZMB		{ return (ZMB); }

{decimal}	{ yylval.val = atoi(yytext);
		  return (NUM);
		}
{hexadecimal}	{ char str[128];
		  yylval.val = (int)(strtol(yytext+1, (char **)NULL, 16));
		  if ((yylval.val & 0xFFFF) == yylval.val)
		    return (NUM);
		  else {
		    sprintf(str, "valor '%s' supera los 16 bits", yytext);
		    show_err(lineno, str);
		    return (ERR);
		  }
		}
{octal}		{ char str[128];
		  yylval.val = (int)(strtol(yytext+1, (char **)NULL, 8));
		  if ((yylval.val & 0xFFFF) == yylval.val)
		    return (NUM);
		  else {
		    sprintf(str, "valor '%s' supera los 16 bits", yytext);
		    show_err(lineno, str);
		    return (ERR);
		  }
		}
{binario}	{ char str[128];
		  yylval.val = (int)(strtol(yytext+1, (char **)NULL, 2));
		  if ((yylval.val & 0xFFFF) == yylval.val)
		    return (NUM);
		  else {
		    sprintf(str, "valor '%s' supera los 16 bits", yytext);
		    show_err(lineno, str);
		    return (ERR);
		  }
		}
{ascii}		{ yylval.val = *(yytext+1);
		  if (yylval.val >= 0x20)
		    return (NUM);
		  else {
		    show_err(lineno, "valor ASCII no permitido");
		    return (ERR);
		  }
		}

{hexerr}	|
{octerr}	|
{binerr}	{ char str[128]; 
		  sprintf(str, "valor '%s' no vlido", yytext);
		  show_err(lineno, str);
		  return (ERR); 
		}

{colon}		{ return (COLON); }
{comma}		{ return (COMMA); }
{inmediato}	{ return (INM); }
{regx}		{ return (RX); }
{regy}		{ return (RY); }

^{simbolo}	{ symrec *s;
		  char str[128];
		  if ((s = getsym(yytext)) == 0)
		    s = putsym(yytext, SYM_ETQ, fname, lineno);
		  else {
		    if (s->type == SYM_UNK) {
		      modsym(s, fname, lineno);
		      // s->line = lineno;
		    }
		    else {
		      sprintf(str, "`%s' ya definido en la lnea %d de `%s'",
			      yytext, s->line, s->fsrc);
		      show_err(lineno, str);
		    }
		  }
		  yylval.sptr = s;
		  return (ETQ);
		}

{simbolo}	{ char    str[128], *j;
		  insrec *i;
		  strcpy(str, yytext);
		  for (j = str; *j != '\0'; j++)
		    *j = toupper(*j);
		  if ((i = getins(str)) != 0) {
		    yylval.iptr = i;
		    return (INS);
		  }
		  else {
		    symrec *s;
		    if ((s = getsym(yytext)) == 0)
		      s = putsym(yytext, SYM_UNK, fname, lineno);
		    yylval.sptr = s;
		    return (SIM);
		  }
		}

\"[^\"]*\"	{ return (STR); }

^\t*\n		|		/* Salta lneas en blanco */
^\*.*\n		lineno++;	/* Salta lneas de comentarios */

\*.*\n		{ show_warn(lineno, "desde '*' asumiendo como comentario");
		  return (FL); 
		}

\n		|		/* Salta lneas */
;.*\n		{ lineno++;	/* Salta comentarios */
		  return (FL);
		}

[ \t\r]		;		/* Salta espacios */

%%
